@@ -3,17 +3,18 @@
 #
 #  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
 
-package Future::IO::Impl::UV;
+package Future::IO::Impl::UV 0.02;
 
-use strict;
+use v5.14;
 use warnings;
 use base qw( Future::IO::ImplBase );
 
-our $VERSION = '0.01';
-
 use UV;
 use UV::Poll;
 use UV::Timer;
+use UV::Signal;
+
+use POSIX ();
 
 __PACKAGE__->APPLY;
 
@@ -25,7 +26,7 @@ C<Future::IO::Impl::UV> - implement C<Future::IO> using C<UV>
 
 This module provides an implementation for L<Future::IO> which uses L<UV>.
 
-There are no additional methods to use in this odule; it simply has to be
+There are no additional methods to use in this module; it simply has to be
 loaded, and it will provide the C<Future::IO> implementation methods:
 
    use Future::IO;
@@ -50,7 +51,7 @@ sub sleep
    return $f;
 }
 
-my %watching_read_by_fileno; # {fileno} => [@futures]
+my %read_futures_by_fileno; # {fileno} => [@futures]
 my %poll_read_by_fileno;
 
 sub ready_for_read
@@ -58,12 +59,12 @@ sub ready_for_read
    shift;
    my ( $fh ) = @_;
 
-   my $watching = $watching_read_by_fileno{ $fh->fileno } //= [];
+   my $futures = $read_futures_by_fileno{ $fh->fileno } //= [];
 
    my $f = Future::IO::Impl::UV::_Future->new;
 
-   my $was = scalar @$watching;
-   push @$watching, $f;
+   my $was = scalar @$futures;
+   push @$futures, $f;
 
    return $f if $was;
 
@@ -72,16 +73,16 @@ sub ready_for_read
    $poll->start( UV::Poll::UV_READABLE, sub {
       my ( $self ) = @_;
 
-      $watching->[0]->done;
-      shift @$watching;
+      $futures->[0]->done;
+      shift @$futures;
 
-      $self->stop if !@$watching;
+      $self->stop if !@$futures;
    });
 
    return $f;
 }
 
-my %watching_write_by_fileno; # {fileno} => [@futures]
+my %write_futures_by_fileno; # {fileno} => [@futures]
 my %poll_write_by_fileno;
 
 sub ready_for_write
@@ -89,12 +90,12 @@ sub ready_for_write
    shift;
    my ( $fh ) = @_;
 
-   my $watching = $watching_write_by_fileno{ $fh->fileno } //= [];
+   my $futures = $write_futures_by_fileno{ $fh->fileno } //= [];
 
    my $f = Future::IO::Impl::UV::_Future->new;
 
-   my $was = scalar @$watching;
-   push @$watching, $f;
+   my $was = scalar @$futures;
+   push @$futures, $f;
 
    return $f if $was;
 
@@ -103,15 +104,56 @@ sub ready_for_write
    $poll->start( UV::Poll::UV_WRITABLE, sub {
       my ( $self ) = @_;
 
-      $watching->[0]->done;
-      shift @$watching;
+      $futures->[0]->done;
+      shift @$futures;
 
-      $self->stop if !@$watching;
+      $self->stop if !@$futures;
    });
 
    return $f;
 }
 
+my $sigchld_watch;
+my %futures_waitpid; # {$pid} => [@futures]
+
+sub waitpid
+{
+   shift;
+   my ( $pid ) = @_;
+
+   # libuv does not currently have a nice way to ask it to watch an existing
+   # PID that it didn't fork/exec itself. All we can do here is ask to be
+   # informed of SIGCHLD and then check if any of the processes we're keeping
+   # an eye on have exited yet. This is kindof sucky, because it means a
+   # linear scan on every signal.
+   #   https://github.com/libuv/libuv/issues/3100
+   $sigchld_watch ||= do {
+      my $w = UV::Signal->new( signal => POSIX::SIGCHLD );
+      $w->start(sub {
+         foreach my $pid ( keys %futures_waitpid ) {
+            next unless waitpid( $pid, POSIX::WNOHANG ) > 0;
+            my $wstatus = $?;
+
+            my $fs = delete $futures_waitpid{$pid};
+            $_->done( $wstatus ) for @$fs;
+         }
+      });
+      $w;
+   };
+
+   my $f = Future::IO::Impl::UV::_Future->new;
+
+   if( waitpid( $pid, POSIX::WNOHANG ) > 0 ) {
+      my $wstatus = $?;
+      $f->done( $wstatus );
+      return $f;
+   }
+
+   push @{ $futures_waitpid{$pid} }, $f;
+
+   return $f;
+}
+
 package Future::IO::Impl::UV::_Future;
 use base qw( Future );
 
